#ifndef FLOW_H
#define FLOW_H

/*
 * flow.h
 * defines the basic classes used by the tcpflow program.
 */


#include <tr1/unordered_map>

class ipaddr {
public:;
    ipaddr(){
	memset(addr,0,sizeof(addr));
    }
    ipaddr(const in_addr_t &a){
	*(in_addr_t *)addr = a;
	memset(addr+4,0,12);
    }
    ipaddr(const uint8_t a[16]){
	memcpy(addr,a,16);
    }

    u_int8_t addr[16];			// holds v4 or v16
    inline bool operator ==(const ipaddr &b) const{
	return 	memcmp(this->addr,b.addr,sizeof(addr))==0;
    };
    inline bool operator <=(const ipaddr &b) const{
	return 	memcmp(this->addr,b.addr,sizeof(addr))<=0;
    };
    inline bool operator >(const ipaddr &b) const{
	return 	memcmp(this->addr,b.addr,sizeof(addr))>0;
    };
    inline bool operator >=(const ipaddr &b) const{
	return 	memcmp(this->addr,b.addr,sizeof(addr))>=0;
    };

    inline bool operator <(const ipaddr &b) const {
	return  memcmp(this->addr,b.addr,sizeof(this->addr))<0;
    }

};

inline std::ostream & operator <<(std::ostream &os,const ipaddr &b)  {
	os << (int)b.addr[0] <<"."<<(int)b.addr[1] << "."
	   << (int)b.addr[2] << "." << (int)b.addr[3];
	return os;
    }

inline bool operator ==(const struct timeval &a,const struct timeval &b) {
    return a.tv_sec==b.tv_sec && a.tv_usec==b.tv_usec;
}

inline bool operator <(const struct timeval &a,const struct timeval &b) {
    return (a.tv_sec<b.tv_sec) || ((a.tv_sec==b.tv_sec) && (a.tv_sec<b.tv_sec));
}

/*
 * describes the TCP flow without the timing information
 */
class flow_addr {
public:
    flow_addr():src(),dst(),sport(0),dport(0),family(0){
    }
    flow_addr(ipaddr s,ipaddr d,uint16_t sp,uint16_t dp,sa_family_t f):
	src(s),dst(d),sport(sp),dport(dp),family(f){
    }
    flow_addr(const flow_addr &f):src(f.src),dst(f.dst),sport(f.sport),dport(f.dport),family(f.family){
    }
    virtual ~flow_addr(){};
    ipaddr	src;		// Source IP address; holds v4 or v6 
    ipaddr	dst;		// Destination IP address; holds v4 or v6 
    uint16_t    sport;		// Source port number 
    uint16_t    dport;		// Destination port number 
    sa_family_t family;		// AF_INET or AF_INET6 */

    uint64_t hash() const {
	if(family==AF_INET){
	    uint64_t s =  *(uint32_t *)src.addr;
	    uint64_t d =  *(uint32_t *)dst.addr;
	    return (s<<32 | d) ^ (d<<32 | s) ^ (sport<<16 | dport);
	} else {
	    uint64_t s0 =  *(uint64_t *)(src.addr+0);
	    uint64_t s8 =  *(uint64_t *)(src.addr+8);
	    uint64_t d0 =  *(uint64_t *)(src.addr+0);
	    uint64_t d8 =  *(uint64_t *)(src.addr+8);
	    return (s0<<32 ^ d0) ^ (d0<<32 ^ s0)
		^ (s8 ^ d8) ^ (sport<<16 | dport);
	}
    }

    inline bool operator ==(const flow_addr &b) const {
	return this->src==b.src &&
	    this->dst==b.dst &&
	    this->sport==b.sport &&
	    this->dport==b.dport &&
	    this->family==b.family;
    }

    inline bool operator <(const flow_addr &b) const {
	if (this->src<b.src) return true;
	if (this->src>b.src) return false;
	if (this->dst<b.dst) return true;
	if (this->dst>b.dst) return false;
	if (this->sport<b.sport) return true;
	if (this->sport>b.sport) return false;
	if (this->dport<b.dport) return true;
	if (this->dport>b.dport) return false;
	if (this->family < b.family) return true;
	if (this->family > b.family) return true;
	return false;    /* they are equal! */
    }
};

inline std::ostream & operator <<(std::ostream &os,const flow_addr &f)  {
    os << "flow[" << f.src << ":" << f.sport << "->" << f.dst << ":" << f.dport << "]";
    return os;
}


/*
 * A flow is a flow_addr that has additional information regarding when it was seen
 * and how many packets were seen. The address is used to locate the flow in the array.
 */
class flow_t : public flow_addr {
public:;
    static int32_t NO_VLAN;			/* vlan flag for no vlan */
    flow_t():id(),vlan(),tstart(),tlast(),packet_count(),connection_count(){};
    flow_t(const flow_addr &flow_addr_,int32_t vlan_,const struct timeval &t1,
	   const struct timeval &t2,uint64_t id_,uint64_t connection_count_):
	flow_addr(flow_addr_),id(id_),vlan(vlan_),tstart(t1),tlast(t2),
	packet_count(0),connection_count(connection_count_){}
    virtual ~flow_t(){};
    uint64_t id;			// flow_counter when this flow was created
    int32_t	vlan;			// vlan interface we observed; -1 means no vlan 
    struct timeval tstart;		// when first seen
    struct timeval tlast;		// when last seen
    uint64_t packet_count;			// packet count
    uint64_t connection_count;	// how many times have we seen a flow with the same addr?
    std::string filename();		// returns filename for a flow
};

/*
 * Convenience class for working with TCP headers
 */
class tcp_header_t {
public:
    tcp_header_t(const u_char *data):
	tcp_header((struct tcphdr *)data){};
    tcp_header_t(const tcp_header_t &b):
	tcp_header(b.tcp_header){}
    tcp_header_t &operator=(const tcp_header_t &that) {
	this->tcp_header = that.tcp_header;
	return *this;
    }

    virtual ~tcp_header_t(){}
    struct tcphdr *tcp_header;
    size_t tcp_header_len(){ return tcp_header->th_off * 4; }
    uint16_t sport() {return ntohs(tcp_header->th_sport);}
    uint16_t dport() {return ntohs(tcp_header->th_dport);}
    tcp_seq  seq()   {return ntohl(tcp_header->th_seq);}
    bool th_fin()    {return tcp_header->th_flags & TH_FIN;}
    bool th_ack()    {return tcp_header->th_flags & TH_ACK;}
    bool th_syn()    {return tcp_header->th_flags & TH_SYN;}
    
};


/* see http://mikecvet.wordpress.com/tag/hashing/ */
typedef struct {
    long operator() (const flow_addr &k) const {return k.hash(); }
} flow_addr_hash;

typedef struct {
    bool operator() (const flow_addr &x, const flow_addr &y) const { return x==y;}
} flow_addr_key_eq;



class tcpip {
public:
    typedef std::tr1::unordered_map<flow_addr,tcpip *,flow_addr_hash,flow_addr_key_eq> flow_map_t; // should be unordered_map
    typedef enum {
	unknown=0,			// unknown direction
	dir_sc,				// server-to-client
	dir_cs				// client-to-server
    } dir_t;
	
private:
    class not_impl: public std::exception {
	virtual const char *what() const throw() {
	    return "copying tcpip objects is not implemented.";
	}
    };
    tcpip(const tcpip &t):flow(),isn(),flow_pathname(),fp(),pos(),pos_max(),last_access(),
			  bytes_printed(),/*fin(),*/
			  finished(),file_exists(),dir(){
	throw new not_impl();
    }
    tcpip &operator=(const tcpip &that) {
	throw new not_impl();
    }
    static uint64_t packet_time;	// monotomically increasing time
    static flow_map_t flow_map;		// the database
public:;
    static xml *xreport;

    /* the flow database */
    static uint64_t flow_counter;	// how many flows have we seen?
    static tcpip *create_tcpip(const flow_addr &flow, int32_t vlan,tcp_seq isn,const timeval &ts,uint64_t connection_count);
    static tcpip *find_tcpip(const flow_addr &flow);
    static void remove_tcpip(const flow_addr &flow);
    static void process_tcp(const struct timeval *ts,const u_char *data, u_int32_t length,
			    const ipaddr &src, const ipaddr &dst,int32_t vlan,sa_family_t family);

    static void flow_map_clear();		// clears out the map

    /* instances - individual tcp/ip flows */
    tcpip(const flow_t &flow_,tcp_seq isn_);    /* constructors */
    virtual ~tcpip();				// destructor
    flow_t  flow;			/* Description of this flow */
    tcp_seq isn;			// Flow's initial sequence number
    std::string flow_pathname;		// path where flow is stored
    FILE     *fp;				// Pointer to file storing this flow's data 
    uint64_t pos;				// Current write position in fp 
    uint64_t pos_max;			// highest pos has gotten
    int last_access;		/* "Time" of last access; used to sort the open flows to figure out which to close */
    uint64_t bytes_printed;		// for -b and -c used together
    bool finished;
    bool file_exists;
    dir_t dir;				// direction of flow

    /* Methods */
    FILE *attempt_fopen(const char *filename);
    FILE *open_file();			// opens this file and returns it
    void close_file();
    void print_packet(const u_char *data, u_int32_t length);
    void store_packet(const u_char *data, u_int32_t length, u_int32_t seq, int syn_set);

};

inline std::ostream & operator << (std::ostream &os,const tcpip::flow_map_t &fm) {
    for(tcpip::flow_map_t::const_iterator it=fm.begin();it!=fm.end();it++){
	//os << "first: " << it->first << " second: " << *it->second << "\n";
    }
    return os;
};

inline std::ostream & operator <<(std::ostream &os,const tcpip &f) {
    //os << "tcpip[" << f.flow << " isn:" << f.isn << " pos:" << f.pos << "]";
    return os;
}




#endif
