/*
 * This file is part of tcpflow by Jeremy Elson <jelson@circlemud.org>
 * Initial Release: 7 April 1999.
 *
 * This source code is under the GNU Public License (GPL).  See
 * LICENSE for details.
 *
 * $Id: tcpip.c,v 1.13 2001/08/24 05:36:14 jelson Exp $
 *
 * $Log: tcpip.c,v $
 * Revision 1.13  2001/08/24 05:36:14  jelson
 * fflush stdout in console print mode, from suggestion of Andreas
 * Schweitzer <andy@physast.uga.edu>, who says "Otherwise, I can't
 * redirect or pipe the console output. At least on FreeBSD. I will check
 * later today if this also cures the same problems I had on OpenBSD."
 *
 * Revision 1.12  2000/12/08 07:32:39  jelson
 * Took out the (broken) support for fgetpos/fsetpos.  Now we always simply
 * use fseek and ftell.
 *
 * Revision 1.11  1999/04/21 01:40:16  jelson
 * DLT_NULL fixes, u_char fixes, additions to configure.in, man page update
 *
 * Revision 1.10  1999/04/20 19:39:19  jelson
 * changes to fix broken localhost (DLT_NULL) handling
 *
 * Revision 1.9  1999/04/14 22:17:40  jelson
 * (re-)fixed checking of fwrite's return value
 *
 * Revision 1.8  1999/04/14 17:59:59  jelson
 * now correctly checking the return value of fwrite
 *
 * Revision 1.7  1999/04/14 03:02:39  jelson
 * added typecasts for portability
 *
 * Revision 1.6  1999/04/13 23:17:56  jelson
 * More portability fixes.  All system header files now conditionally
 * included from sysdep.h.
 *
 * Integrated patch from Johnny Tevessen <j.tevessen@gmx.net> for Linux
 * systems still using libc5.
 *
 * Revision 1.5  1999/04/13 01:38:15  jelson
 * Added portability features with 'automake' and 'autoconf'.  Added AUTHORS,
 * NEWS, README, etc files (currently empty) to conform to GNU standards.
 *
 * Various portability fixes, including the FGETPOS/FSETPOS macros; detection
 * of header files using autoconf; restructuring of debugging code to not
 * need vsnprintf.
 *
 */

#include "tcpflow.h"
#include <netinet/ip6.h>		/*  SLG */
#include <iostream>
#include <sstream>

xml *tcpip::xreport =0;

/****************************************************************
 *** Legacy flow database
 ****************************************************************/
static int max_fds=0;
static int next_slot=-1;
static tcpip **fd_ring;

/* Initialize our structures */
void init_tcpip()
{
    /* Find out how many files we can have open safely...subtract 4 for
     * stdin, stdout, stderr, and the packet filter; one for breathing
     * room (we open new files before closing old ones), and one more to
     * be safe.
     */
    max_fds = get_max_fds() - NUM_RESERVED_FDS;

    fd_ring = (tcpip **)calloc(sizeof(tcpip *), max_fds);
    next_slot = -1;
}


/****************************************************************
 *** Instance methods
 ****************************************************************/

/**
 * Close all of the flows in the fd_ring
 */
void flow_close_all()
{
    int i;
    for(i=0;i<max_fds;i++){
	if(fd_ring[i]){
	    fd_ring[i]->close_file();
	    fd_ring[i] = 0;
	}
    }
}



/* This comparison function puts flows first in the array, and nulls
 * last.  Within the flows, they are ordered from least recently
 * accessed at the front, and most recently accessed at the end. */
int tcpip_compare(const void *a, const void *b)
{
    tcpip **x = (tcpip **)a;
    tcpip **y = (tcpip **)b;

    if (*x == NULL && *y == NULL) return 0;
    if (*x == NULL) return 1;
    if (*y == NULL) return -1;
    return ((*x)->last_access - (*y)->last_access);
}


/* Sort FDs in the fd_table according to the comparison function (see
 * comment above) */
void sort_fds()
{
    qsort(fd_ring, max_fds, sizeof(tcpip *), tcpip_compare);
}


/* We need to reduce the size of the fd ring by one FD.  We will
 * sort the FD ring, close the oldest (i.e. first) file descriptor,
 * shift everything down by one, and set max_fds to reflect the new
 * size. */
void contract_fd_ring()
{
    /* sort */
    sort_fds();

    /* make sure we're sane */
    if (fd_ring[0] == NULL) {
	die("we seem to be completely out of file descriptors");
    }

    /* close the oldest FD */
    fd_ring[0]->close_file();

    /* shift everything forward by one and count */
    int i=1;
    for (i = 1; i < max_fds && fd_ring[i] != NULL; i++)
	fd_ring[i-1] = fd_ring[i];

    /* remember that the ring is smaller now */
    max_fds = i-1;

    /* start at 0 (by setting to -1, since we're going to increment it) */
    next_slot = -1;
}

/* Added by SLG:
 * Remove a given tcpip from the ring.
 * Doesn't make the ring smaller; ring will be shortened on sort.
 */
void remove_fd_ring(tcpip *s)
{
    /* find the element */
    for(int i=0;i<max_fds; i++){
	if(fd_ring[i]==s){
	    fd_ring[i]->close_file();
	    fd_ring[i] = 0;
	    return;
	}
    }
}



/****************************************************************
 *** THE TCPIP FLOW DATABASE
 ****************************************************************/

uint64_t tcpip::flow_counter = 0;
tcpip::flow_map_t tcpip::flow_map;
uint64_t tcpip::packet_time = 0;

/* Find previously a previously created flow state in the database.
 */
tcpip *tcpip::find_tcpip(const flow_addr &flow)
{
    flow_map_t::const_iterator it = flow_map.find(flow);
    if (it==flow_map.end()){
	return NULL; // flow not found
    }
    return it->second;
}

/* Create a new flow state structure for a given flow.
 * Puts the flow in the map.
 * Returns a pointer to the new state.
 * Do we really want to keep pointers to tcp structures in the map, rather than
 * the structures themselves?
 */

tcpip *tcpip::create_tcpip(const flow_addr &flowa, int32_t vlan,tcp_seq isn,
			   const timeval &ts,uint64_t connection_count)
{
    /* create space for the new state */
    flow_t flow(flowa,vlan,ts,ts,tcpip::flow_counter++,connection_count);

    tcpip *new_tcpip = new tcpip(flow,isn);
    new_tcpip->last_access = packet_time++;
    DEBUG(5) ("%s: new flow", new_tcpip->flow_pathname.c_str());
    flow_map[flow] = new_tcpip;
    return new_tcpip;
}

void tcpip::remove_tcpip(const flow_addr &flow)
{
    flow_map_t::iterator it = flow_map.find(flow);
    if(it!=flow_map.end()){
	delete it->second;
	flow_map.erase(it);
    }
}

void tcpip::flow_map_clear()
{
    for(tcpip::flow_map_t::iterator it=flow_map.begin();it!=flow_map.end();it++){
	delete it->second;
    }
    flow_map.clear();
}

tcpip::tcpip(const flow_t &flow_,tcp_seq isn_):
    flow(flow_),isn(isn_),flow_pathname(),fp(0),pos(0),pos_max(0),
    last_access(0),bytes_printed(0),finished(0),file_exists(0),dir(unknown)
{
    if(opt_output_enabled){
	if(outdir=="."){
	    flow_pathname = flow.filename();
	} else {
	    flow_pathname = outdir + std::string("/") + flow.filename();
	}
    }
}


tcpip::~tcpip()
{
    static std::string filesize_string("filesize");

    if(xreport){
	xreport->push("fileobject");
	if(flow_pathname.size()) xreport->xmlout("filename",flow_pathname);
	xreport->xmlout(filesize_string,pos_max);
	
	std::stringstream ss;
	ss << "startime='" << xml::to8601(flow.tstart) << "' ";
	ss << "endtime='"  << xml::to8601(flow.tlast)  << "' ";
	ss << "src_ipn='"  << flow.src << "' ";
	ss << "dst_ipn='"  << flow.dst << "' ";
	ss << "packets='"  << flow.packet_count << "' ";
	ss << "srcport='"  << flow.sport << "' ";
	ss << "dstport='"  << flow.dport << "' ";
	ss << "family='"   << flow.family << "' ";
	xreport->xmlout("tcpflow","",ss.str(),false);
	xreport->pop();
    }
}


FILE *tcpip::attempt_fopen(const char *filename)
{
    /* If we've opened this file already, reopen it.  Otherwise create a
     * new file.  We purposefully overwrite files from previous runs of
     * the program.
     */
    if (file_exists) {
	DEBUG(5) ("%s: re-opening output file", filename);
	fp = fopen(filename, "r+");
    } else {
	DEBUG(5) ("%s: opening new output file", filename);
	fp = fopen(filename, "w");
    }

    return fp;
}


FILE *tcpip::open_file()
{
    int done;

    /* This shouldn't be called if the file is already open */
    if (fp) {
	DEBUG(20) ("huh -- trying to open already open file!");
	return fp;
    }

    /* Now try and open the file */
    do {
	if (attempt_fopen(flow_pathname.c_str()) != NULL) {
	    /* open succeeded... great */
	    done = 1;
	} else {
	    if (errno == ENFILE || errno == EMFILE) {
		/* open failed because too many files are open... close one
		 * and try again
		 */
		contract_fd_ring();
		DEBUG(5) ("too many open files -- contracting FD ring to %d", max_fds);
		done = 0;
	    } else {
		/* open failed for some other reason... give up */
		done = 1;
	    }
	}
    } while (!done);

    /* If the file isn't open at this point, there's a problem */
    if (fp == NULL) {
	/* we had some problem opening the file -- set FINISHED so we
	 * don't keep trying over and over again to reopen it
	 */
	finished = true;
	perror(flow_pathname.c_str());
	return NULL;
    }

    /* Now we decide which FD slot we use, and close the file that's
     * there (if any).  Note that even if tcpip is not NULL, its
     * associated file pointer may already be closed.  Note well that we
     * DO NOT free the state that we find in our slot; the state stays
     * around forever (pointed to by the hash table).  This table only
     * keeps a pointer to state structures that have open files so that
     * we can close them later.
     *
     * We are putting the close after the open so that we don't bother
     * closing files if the open fails.  (For this, we pay a price of
     * needing to keep a spare, idle FD around.) */

    if (++next_slot == max_fds) {
	/* take this opportunity to sort from oldest to newest --
	 * optimally we'd like to do this before every close, but that
	 * might take too long. */
	sort_fds();
	next_slot = 0;
    }

    /* close the next one in line */
    if (fd_ring[next_slot] != NULL){
	fd_ring[next_slot]->close_file();
    }

    /* put ourslves in its place */
    fd_ring[next_slot] = this;

    /* set flags and remember where in the file we are */
    file_exists = true;
    pos = ftell(fp);
    return fp;
}



/* Closes the file belonging to a flow -- returns 1 if a file was
 * actually closed, 0 otherwise (if it was already closed) */
void tcpip::close_file()
{
    if (fp){
	struct timeval times[2];
	times[0] = flow.tstart;
	times[1] = flow.tstart;

	DEBUG(5) ("%s: closing file", flow_pathname.c_str());
	/* close the file and remember that it's closed */
	fflush(fp);		/* flush the file */
	if(futimes(fileno(fp),times)){
	    perror("futimes");
	}
	fclose(fp);
	fp = NULL;
	pos = 0;
    }
    //return 1;
}






/*************************************************************************/

/* convert all non-printable characters to '.' (period).  
 */
static u_char *do_strip_nonprint(const u_char *data, u_char *buf,u_int32_t length)
{
    u_char *write_ptr = buf;
    while (length) {
	if (isprint(*data) || (*data == '\n') || (*data == '\r'))
	    *write_ptr = *data;
	else
	    *write_ptr = '.';
	write_ptr++;
	data++;
	length--;
    }
    return buf;
}



/* print the contents of this packet to the console */
void tcpip::print_packet(const u_char *data, u_int32_t length)
{
    /* green, blue, read */
    const char *color[3] = { "\033[0;32m", "\033[0;34m", "\033[0;31m" };

    if(bytes_per_flow>0){
	if(bytes_printed>bytes_per_flow) return; /* too much has been printed */
	if(length > bytes_per_flow - bytes_printed){
	    length = bytes_per_flow - bytes_printed; /* can only output this much */
	    if(length==0) return;
	}
    }

#ifdef HAVE_PTHREAD
    if(semlock){
	if(sem_wait(semlock)){
	    fprintf(stderr,"%s: attempt to acquire semaphore failed: %s\n",progname,strerror(errno));
	    exit(1);
	}
    }
#endif

    if (use_color) {
	fputs(dir==dir_cs ? color[1] : color[2], stdout);
    }

    if (suppress_header == 0) {
	printf("%s: ", flow_pathname.c_str());
    }

    fwrite(data, length, 1, stdout);
    bytes_printed += length;

    if (use_color) printf("\033[0m");

    putchar('\n');
    fflush(stdout);

#ifdef HAVE_PTHREAD
    if(semlock){
	if(sem_post(semlock)){
	    fprintf(stderr,"%s: attempt to post semaphore failed: %s\n",progname,strerror(errno));
	    exit(1);
	}
    }
#endif
}


/* store the contents of this packet to its place in its file */
void tcpip::store_packet(const u_char *data, u_int32_t length, u_int32_t seq, int syn_set)
{
    /* If we got a SYN reset the sequence number */
    if (syn_set) {
	DEBUG(50) ("resetting isn due to extra SYN");
	isn = seq - pos +1;
    }

    /* if we're done collecting for this flow, return now */
    if (finished) return;

    /* calculate the offset into this flow -- should handle seq num
     * wrapping correctly because tcp_seq is the right size */
    tcp_seq offset = seq - isn;

    /* I want to guard against receiving a packet with a sequence number
     * slightly less than what we consider the ISN to be; the max
     * (though admittedly non-scaled) window of 64K should be enough */
    if (offset >= 0xffff0000) {
	DEBUG(2) ("dropped packet with seq < isn on %s", flow_pathname.c_str());
	//std::cerr << "dropped packet with seq < isn on " << flow_pathname << "\n";
	return;
    }

    /* reject this packet if it falls entirely outside of the range of
     * bytes we want to receive for the flow */
    if (bytes_per_flow && (offset > bytes_per_flow))
	return;

    /* reduce length if it goes beyond the number of bytes per flow */
    if (bytes_per_flow && (offset + length > bytes_per_flow)) {
	finished = true;
	length = bytes_per_flow - offset;
    }

    if (opt_output_enabled){
	/* if we don't have a file open for this flow, try to open it.
	 * return if the open fails.  Note that we don't have to explicitly
	 * save the return value because open_file() puts the file pointer
	 * into the structure for us. */
	if (fp == NULL) {
	    if (open_file() == NULL) {
		return;
	    }
	}
	
	/* if we're not at the correct point in the file, seek there */
	if (offset != pos) {
	    fseek(fp, offset, SEEK_SET);
	}
	
	/* write the data into the file */
	DEBUG(25) ("%s: writing %ld bytes @%ld", flow_pathname.c_str(),
		   (long) length, (long) offset);
	
	if (fwrite(data, length, 1, fp) != 1) {
	    /* sigh... this should be a nice, plain DEBUG statement that
	     * passes strerrror() as an argument, but SunOS 4.1.3 doesn't seem
	     * to have strerror. */
	    if (debug_level >= 1) {
		DEBUG(1) ("write to %s failed: ", flow_pathname.c_str());
		perror("");
	    }
	}
	fflush(fp);
    }

    /* remember the position for next time */
    pos = offset + length;
    if(pos>pos_max) pos_max = pos;

    if (finished) {
	DEBUG(5) ("%s: stopping capture", flow_pathname.c_str());
	close_file();
    }
}

/*
 * Called to processes a tcp packet
 */

void tcpip::process_tcp(const struct timeval *ts,const u_char *data, u_int32_t length,
			const ipaddr &src, const ipaddr &dst,int32_t vlan,sa_family_t family)
{
    struct tcphdr *tcp_header = (struct tcphdr *) data;
    u_int tcp_header_len;
    tcp_seq seq;

    if (length < sizeof(struct tcphdr)) {
	DEBUG(6) ("received truncated TCP segment!");
	return;
    }

    /* calculate the total length of the TCP header including options */
    tcp_header_len = tcp_header->th_off * 4;

    /* fill in the flow_addr structure with info that identifies this flow */
    flow_addr this_flow(src,dst,ntohs(tcp_header->th_sport),ntohs(tcp_header->th_dport),family);

    seq = ntohl(tcp_header->th_seq);
    int syn_set = IS_SET(tcp_header->th_flags, TH_SYN);
    /* recalculate the beginning of data and its length, moving past the
     * TCP header
     */
    data   += tcp_header_len;
    length -= tcp_header_len;

    /* see if we have state about this flow; if not, create it (from Debian patch 10) */
    tcpip *state = tcpip::find_tcpip(this_flow);
    uint64_t connection_count = 0;
    if(state){
	/* If offset will be too much, throw away this_flow and create a new one */
	tcp_seq isn2 = state->isn;		// local copy
	if(syn_set){
	    isn2 = seq - state->pos + 1;
	}
	tcp_seq offset = seq - isn2;
	if(offset != state->pos) {
	    /* there will be a seek to pos. */
	    if((offset > state->pos)
	       && (offset > state->pos_max)
	       && (length > 0)
	       && (bytes_per_flow==0) 
	       && ((uint64_t)offset > (uint64_t)state->pos +  min_skip)){
		//std::cerr << "state->pos=" << state->pos << " state->pos_max="
		//<< state->pos_max << " offset=" << offset << " min_skip="
		//<< min_skip << "  this_flow=" << this_flow << "\n";
		connection_count = state->flow.connection_count+1;
		tcpip::remove_tcpip(this_flow);
		state = 0;
	    }
	}
    }
    if (state==NULL){
	state = tcpip::create_tcpip(this_flow, vlan, seq, *ts,connection_count);
    }

    state->flow.packet_count++;

    /* Handle empty packets (from Debian patch 10) */
    if (length == 0) {
	/* examine TCP flags for initial TCP handshake segments:
	 * - SYN means that the flow is a client -> server flow
	 * - SYN/ACK means that the flow is a server -> client flow.
	 */
	if ((state->isn - seq) == 0) {
	    if (IS_SET(tcp_header->th_flags, TH_SYN) && IS_SET(tcp_header->th_flags, TH_ACK)) {
		state->dir = dir_sc;
		DEBUG(50) ("packet is handshake SYN/ACK");
		/* If the SYN flag is set the first data byte is offset by one,
		 * account for it (note: if we're here we have just created
		 * state, so it's safe to change isn).
		 */
		state->isn++;
	    } else if (IS_SET(tcp_header->th_flags, TH_SYN)) {
		state->dir = dir_cs;
		DEBUG(50) ("packet is handshake SYN");
		state->isn++;
	    }
	}
	DEBUG(50) ("got TCP segment with no data");
    }

    if (length>0){
	/* strip nonprintable characters if necessary */
	u_char newbuf[SNAPLEN];
	if (strip_nonprint){
	    data = do_strip_nonprint(data, newbuf,length);
	}
	
	/* store or print the output */
	if (console_only) {
	    state->print_packet(data, length);
	} else {
	    state->store_packet(data, length, seq, syn_set);
	}
    }

    /* Finally, if there is a FIN, then kill this TCP connection*/
    if (IS_SET(tcp_header->th_flags, TH_FIN)){
	if(opt_no_purge==false){
	    DEBUG(50)("packet is FIN; closing connection");
	    remove_fd_ring(state);		// remove the state from the ring
	    remove_tcpip(this_flow);	// take it out of the map
	}
    }
}





/* This is called when we receive an IPv4 datagram.  We make sure that
 * it's valid and contains a TCP segment; if so, we pass it to
 * process_tcp() for further processing.
 *
 * Note: we currently don't know how to handle IP fragments. */
void process_ip4(const struct timeval *ts,const u_char *data, u_int32_t caplen,int32_t vlan)
{
    const struct ip *ip_header = (struct ip *) data;
    u_int ip_header_len;
    u_int ip_total_len;

    /* make sure that the packet is at least as long as the min IP header */
    if (caplen < sizeof(struct ip)) {
	DEBUG(6) ("received truncated IP datagram!");
	return;
    }

    /* for now we're only looking for TCP; throw away everything else */
    if (ip_header->ip_p != IPPROTO_TCP) {
	DEBUG(50) ("got non-TCP frame -- IP proto %d", ip_header->ip_p);
	return;
    }

    /* check and see if we got everything.  NOTE: we must use
     * ip_total_len after this, because we may have captured bytes
     * beyond the end of the packet (e.g. ethernet padding).
     */
    ip_total_len = ntohs(ip_header->ip_len);
    if (caplen < ip_total_len) {
	DEBUG(6) ("warning: captured only %ld bytes of %ld-byte IP datagram",
		  (long) caplen, (long) ip_total_len);
    }

    /* XXX - throw away everything but fragment 0; this version doesn't
     * know how to do fragment reassembly.
     */
    if (ntohs(ip_header->ip_off) & 0x1fff) {
	DEBUG(2) ("warning: throwing away IP fragment from X to X");
	return;
    }

    /* figure out where the IP header ends */
    ip_header_len = ip_header->ip_hl * 4;

    /* make sure there's some data */
    if (ip_header_len > ip_total_len) {
	DEBUG(6) ("received truncated IP datagram!");
	return;
    }

    /* do TCP processing, faking an ipv6 address  */
    tcpip::process_tcp(ts,data + ip_header_len, ip_total_len - ip_header_len,
		ipaddr(ip_header->ip_src.s_addr),
		ipaddr(ip_header->ip_dst.s_addr),
		vlan,AF_INET);
}


/* This is called when we receive an IPv6 datagram.
 *
 * Note: we don't support IPv6 extended headers
 */


void process_ip6(const struct timeval *ts,const u_char *data,
		 const u_int32_t caplen, const int32_t vlan)
{
    const struct ip6_hdr *ip_header = (struct ip6_hdr *) data;
    u_int16_t ip_payload_len;

    /* make sure that the packet is at least as long as the IPv6 header */
    if (caplen < sizeof(struct ip6_hdr)) {
	DEBUG(6) ("received truncated IPv6 datagram!");
	return;
    }


    /* for now we're only looking for TCP; throw away everything else */
    if (ip_header->ip6_nxt != IPPROTO_TCP) {
	DEBUG(50) ("got non-TCP frame -- IP proto %d", ip_header->ip6_nxt);
	return;
    }

    ip_payload_len = ntohs(ip_header->ip6_plen);

    /* make sure there's some data */
    if (ip_payload_len == 0) {
	DEBUG(6) ("received truncated IP datagram!");
	return;
    }

    /* do TCP processing */

    tcpip::process_tcp(ts,
		data + sizeof(struct ip6_hdr),
		ip_payload_len,
		ipaddr(ip_header->ip6_src.s6_addr),
		ipaddr(ip_header->ip6_dst.s6_addr),
		vlan,AF_INET6);
}



/* This is called when we receive an IPv4 or IPv6 datagram.
 * This function calls process_ip4 or process_ip6
 */

void process_ip(const struct timeval *ts,const u_char *data, u_int32_t caplen,int32_t vlan)
{
    const struct ip *ip_header = (struct ip *) data;
    if (caplen < sizeof(struct ip)) {
	DEBUG(6) ("can't determine IP datagram version!");
	return;
    }

    if(ip_header->ip_v == 6) {
	process_ip6(ts,data, caplen,vlan);
    } else {
	process_ip4(ts,data, caplen,vlan);
    }
}
 
 
